local mod = ...



local rangeTower = 100



-- ************************************************************************************************
-- Component: Fire --------------------------------------------------------------------------
-- ************************************************************************************************


-- Component
local COMP_FIRE = {
	TypeName = "COMP_FIRE",
	ParentType = "COMPONENT",
	Properties = {
		{ Name = "InitFlag", Type = "boolean", Default = false, Flags = { "SAVE_GAME" } },
		{ Name = "Position", Type = "vec3f", Default = nil, Flags = { "SAVE_GAME" } },
		{ Name = "BeginEmitterCount", Type = "integer", Default = 1, Flags = { "SAVE_GAME" } },
        { Name = "ObjectList", Type = "list<string>", Default = nil, Flags = { "SAVE_GAME" } },
        { Name = "StateList", Type = "list<integer>", Default = nil, Flags = { "SAVE_GAME" } },
        { Name = "DistList", Type = "list<float>", Default = nil, Flags = { "SAVE_GAME" } },
        { Name = "ClearedList", Type = "list<boolean>", Default = nil, Flags = { "SAVE_GAME" } },
        { Name = "TimerList", Type = "list<float>", Default = nil, Flags = { "SAVE_GAME" } },
        { Name = "EmitterList", Type = "list<string>", Default = nil, Flags = { "SAVE_GAME" } },
        { Name = "EachSpreadDistance", Type = "float", Default = 5, Flags = { "SAVE_GAME" } },
		{ Name = "MaxSpreadDistance", Type = "float", Default = 20, Flags = { "SAVE_GAME" } },
		{ Name = "TimerTotal", Type = "float", Default = 50, Flags = { "SAVE_GAME" } },
		{ Name = "TimerSpread", Type = "float", Default = 20, Flags = { "SAVE_GAME" } },
		{ Name = "TimerSink", Type = "float", Default = 10, Flags = { "SAVE_GAME" } },
		{ Name = "ChildEmitterLimit", Type = "integer", Default = 5, Flags = { "SAVE_GAME" } },
		{ Name = "SpeedFall", Type = "float", Default = 2, Flags = { "SAVE_GAME" } },
		{ Name = "WaterList", Type = "list<string>", Default = nil, Flags = { "SAVE_GAME" } },
		{ Name = "ActiveCount", Type = "integer", Default = 0, Flags = { "SAVE_GAME" } },
		{ Name = "WaterCount", Type = "integer", Default = 0, Flags = { "SAVE_GAME" } },
		{ Name = "TimerDaysSelfDestruct", Type = "integer", Default = 7, Flags = { "SAVE_GAME" } },
    }
}



-- Init ----------------------------------------------------------------------------------------------
function COMP_FIRE:init()


    -- set up daily event
    local compMainGameLoop = self:getLevel():find("COMP_MAIN_GAME_LOOP")
    event.register(self, compMainGameLoop.ON_NEW_DAY, function()

        -- Self destruct after n days have elapsed from the fire
        if self.TimerDaysSelfDestruct <= 0 then
            -- mod:log("COMP_FIRE: deleting the original fire object.")
            self:getOwner():destroy()
        else
            self.TimerDaysSelfDestruct = self.TimerDaysSelfDestruct - 1
        end

    end) -- end daily event


    -- very first run
    if not self.InitFlag then


        -- Set variables
        local objFire = self:getOwner()
        local posFire = objFire:getGlobalPosition()
        self.Position = posFire


        -- Record the fire in the controller, and perform other meta-tasks
        local compControl = self:getLevel():getComponentManager("COMP_FIRE_CONTROLLER"):getFirstEnabled()
        if compControl then
            compControl:registerFire(objFire)
        end


        -- Record the wells within range

        -- Vanilla well
        self:recordWater(posFire, "WellPart1")

        -- Mod: Water sources and Pagan Worship
        self:recordWater(posFire, "Spring1Part")
        self:recordWater(posFire, "WellPart")

        -- Mod: Forest Folk
        self:recordWater(posFire, "Water1Part")

        -- Mod: Archipelago map
        self:recordWater(posFire, "BUILDING_ARCHIPELAGO_WATER_WORKPLACE_NAME")
        self:recordWater(posFire, "BUILDING_PART_ARCHIPELAGO_WATER_NAME")


        -- Save a list of trees and buildings for spreading the fire
        -- this reduces the performance impact of searching for nearby objects, and limits the total spread of the fire

        -- Building parts
        self:getLevel():getComponentManager("COMP_BUILDING_PART"):getAllEnabledComponent():forEach(function(comp)

            local obj = comp:getOwner()
            if comp:isVisibleWhenBuilt() and comp:getIsBuilt() and obj.Active and obj.Name ~= "TempVillageCenterPart" and obj.Name ~= "VillageCenter" and obj.Name ~= "OutlawCamp.Part" then
                self:recordObject(obj)
            end

        end)


        -- Trees
        self:getLevel():getComponentManager("COMP_TREE"):getAllEnabledComponent():forEach(function(comp)

            local obj = comp:getOwner()
            if comp.IsActive and obj.Active then
                self:recordObject(obj)
            end

        end)


        -- Crops
        self:getLevel():getComponentManager("COMP_CROP_FIELD_ELEMENT"):getAllEnabledComponent():forEach(function(comp)

            local obj = comp:getOwner()
            if obj.Active then
                self:recordObject(obj)
            end

        end)


        -- Stumps (mod)
        if (foundation.isModLoaded("59f226a2-a363-4804-bf91-668dec627b70")) then

            self:getLevel():getComponentManager("COMP_STUMP"):getAllEnabledComponent():forEach(function(comp)

                local obj = comp:getOwner()
                if obj.Active then
                    self:recordObject(obj)
                end

            end)

        end


        -- Rubbish (mod)
        if (foundation.isModLoaded("f1b1db50-e429-4901-9c3b-1bd4a66f876f")) then
            self:getLevel():getComponentManager("COMP_RUBBISH"):getAllEnabledComponent():forEach(function(comp)

                local obj = comp:getOwner()
                if obj.Active then
                    self:recordObject(obj)
                end

            end)
        end


        -- Start the first emitter with the closest object
        local distNearest = 1000
        local objNearest = nil
        local indNearest = nil
        for p = 1, #self.DistList do
            if self.DistList[p] < distNearest then
                objNearest = self.ObjectList[p]
                distNearest = self.DistList[p]
                indNearest = p
            end
        end

        if objNearest then
            local obj = self:getLevel():find(objNearest)
            if obj and obj.Active then
                self:setObjectOnFire(obj, indNearest)
            end
        end


        -- Set an additional delay if there is a Fire Watchtower nearby
        self:getLevel():getComponentManager("COMP_FIRE_TOWER"):getAllEnabledComponent():forEach(function(compTower)

            local objTower = compTower:getOwner()
            if objTower.Active then

                -- Find out if the tower is within range
                local posTower = objTower:getGlobalPosition()
                local dist = math.sqrt( (self.Position[1] - posTower[1])^2 + (self.Position[3] - posTower[3])^2 )
                if (dist <= rangeTower) then

                    -- Find out if the tower is manned
                    local compWork = objTower:findFirstObjectWithComponentDown("COMP_WORKPLACE")
                    if compWork then

                        local listWorker = compWork.AssignedWorkerList
                        if listWorker then

                            local countWorker = 0
                            for n = 1, #listWorker do
                                countWorker = countWorker + 1
                            end

                            if countWorker > 0 then

                                -- Lengthen the timer on the first object (the fire itself), which will delay its spread
                                self.FloatList[1].Timer = self.FloatList[1].Timer + self.TimerTotal

                            end

                        end -- listWorker exists
                    end -- compWork exists
                end -- distance is within range
            end -- tower is active
        end) -- loop through all towers


        -- Don't repeat init on reload
        self.InitFlag = true


        -- mod:log(#self.ObjectList .. " objects recorded within maximum Fire radius.")


    end -- end initflag not set

end



-- Record object
function COMP_FIRE:recordObject(obj)

    local posObj = obj:getGlobalPosition()
    local dist = self:getNearbyObject(posObj, self.Position, self.MaxSpreadDistance)
    if dist then
        -- table.insert(self.ObjectList, {Object = obj:getId(), Dist = dist, State = 0, Cleared = 0, Timer = self.TimerTotal + math.random(-3, 3) })
        table.insert(self.ObjectList, obj:getId())
        table.insert(self.StateList, 0)
        table.insert(self.DistList, dist)
        table.insert(self.ClearedList, false)
        table.insert(self.TimerList, self.TimerTotal + math.random(-3, 3))
        table.insert(self.EmitterList, nil)
    end

end



-- Record water source
function COMP_FIRE:recordWater(posFire, str)

    local listWater = {}
    self:getLevel():find(str, listWater)

    if listWater then
        for i = 1, #listWater do

            local objWater = listWater[i]
            local posWater = objWater:getGlobalPosition()
            local dist = math.sqrt( (posFire[1] - posWater[1])^2 + (posFire[3] - posWater[3])^2 )

            if dist <= rangeTower then
                table.insert(self.WaterList, listWater[i]:getId())
            end

        end
    end
end



-- Find out if an object is nearby or not
function COMP_FIRE:getNearbyObject(posObj, posFire, maxDist)

    local distX = math.abs(posFire[1] - posObj[1])
    local distY = math.abs(posFire[3] - posObj[3])
    -- local dist = math.sqrt( (posFire[1] - posObj[1])^2 + (posFire[3] - posObj[3])^2 )

    if (distX < maxDist) and (distY < maxDist) then

        local dist = 0.5 * (distX + distY + math.max(distX, distY)) -- this is an approximate distance but is faster than using square roots

        if dist < maxDist then -- spread within a short distance
            return dist
        else
            return false
        end

    end

end



-- Update ------------------------------------------------------------------------------------------------------
function COMP_FIRE:update()

    -- ObjectList states:
    -- 0 = not on fire
    -- 1 = on fire but hasn't spread yet
    -- 2 = on fire and has already been spread
    -- 3 = destroyed

    -- get time elapsed since last frame update
    local dt = self:getLevel():getDeltaTime()

    if dt > 0 and dt < 0.20 then -- FPS limiter; do not take any action if FPS is too low; also, do not process if the fire is already put out

        -- for i = 1, #self.ObjectList do
            -- mod:log(tostring(self.ObjectList[i].Object) .. ", type " .. tostring(self.ObjectList[i].type) .. ", state " .. tostring(self.ObjectList[i].State) .. ", timer " .. tostring(self.ObjectList[i].Timer))
        -- end


        -- Loop through all objects

        self.ActiveCount = 0

        for i = 1, #self.ObjectList do

            -- -- Testing
            -- if i == 10 then
                -- mod:log(i .. " " .. tostring(self.ObjectList[i]) .. " " .. tostring(self.StateList[i]) .. " " .. tostring(self.DistList[i]) .. " " .. tostring(self.ClearedList[i]) .. " " .. tostring(self.TimerList[i]))
            -- end


            -- Only pick objects that are actively burning
            if self.StateList[i] > 0 and self.StateList[i] < 3 then

                self.ActiveCount = self.ActiveCount + 1
                local obj = self:getLevel():find(self.ObjectList[i])

                if obj then

                    -- Update timer if the object has already started the fire lifecycle
                    self.TimerList[i] = self.TimerList[i] - dt
                    -- mod:log("self.TimerList[i] " .. i .. " " .. self.TimerList[i])


                    -- Put out a fire that hasn't spread yet and there is water available; this resets the object timer and status
                    if self.StateList[i] == 1 and self.WaterCount > 0 then

                        -- Set new status
                        if obj.Name == "FireStarter.Part" then -- will never burn again, begin end of lifecycle
                            self.StateList[i] = 1
                            self.TimerList[i] = 0
                            -- self.TimerList[i] = self.TimerSink

                        else -- can be re-burned by another nearby object
                            self.StateList[i] = 0
                            self.TimerList[i] = self.TimerTotal + math.random(-3, 3)
                            self.ActiveCount = self.ActiveCount - 1

                        end

                        -- Turn off emitter
                        if self.EmitterList[i] then
                            local objEmitter = self:getLevel():find(self.EmitterList[i])
                            if objEmitter then
                                objEmitter:forEachChild(function(child)
                                    local compEmitter = child:getComponent("COMP_PARTICLE_EMITTER")
                                    if compEmitter then
                                        compEmitter.IsEmitting = false
                                    end
                                end)
                            end
                        end

                        -- Decrement water
                        self.WaterCount = self.WaterCount - 1


                    -- Spread fire to nearby objects from objects that are actively burning
                    elseif self.StateList[i] == 1 and self.TimerList[i] < self.TimerSpread then

                        -- Find nearby objects
                        local posObj = obj:getGlobalPosition()

                        for n = 1, #self.ObjectList do

                            -- mod:log("   " .. n .. " " .. tostring(self.ObjectList[i]) .. " " .. tostring(self.StateList[i]) .. " " .. tostring(self.DistList[i]) .. " " .. tostring(self.ObjectList.Cleared[i]) .. " " .. tostring(self.TimerList[i]))

                            -- local self.ObjectList[i]N = self.ObjectList[n]
                            -- local self.StateList[i]N = self.StateList[n]
                            -- local self.DistList[i]N = self.DistList[n]

                            -- Check only for items that aren't on fire yet
                            if self.StateList[n] == 0 and n ~= i then

                                -- Check only within the band of distances near the first point, to cut down the possibilities
                                if (self.DistList[n] <= (self.DistList[i] + self.EachSpreadDistance)) then

                                    local objN = self:getLevel():find(self.ObjectList[n])

                                    -- Only proceed if the object still exists
                                    if objN then

                                        -- Only proceed if the object is active (visible)
                                        if objN.Active then

                                            -- Only proceed if the object is nearby
                                            local posObjN = objN:getGlobalPosition()
                                            local dist = self:getNearbyObject(posObjN, posObj, self.EachSpreadDistance)
                                            if dist then

                                                -- Set object on fire
                                                self:setObjectOnFire(objN, n)

                                            end -- n object is within exact range

                                        end -- n object is active

                                    end -- n object exists

                                end -- n object is within the range band

                            end -- n state is zero

                        end -- loop through all n objects

                        -- New state: on fire and has already spread, cannot be stopped from burning
                        self.StateList[i] = 2

                        -- Disable workplaces and generators
                        obj:forEachComponent(function(comp)
                            if comp:getComponentType() == "COMP_WORKPLACE" or comp:getComponentType() == "COMP_RESOURCE_GENERATOR" or comp:getComponentType() == "CUSTOM_DECORATION_V2_DECORATION_COMP" or comp:getComponentType() == "CUSTOM_DECORATION_V3_DECORATION_COMP" then
                                obj:removeComponent(comp)
                            end
                        end)


                    -- Object is on fire, has spread and is sinking
                    elseif (self.StateList[i] == 2) and (self.TimerList[i] >= 0) and (self.TimerList[i] < self.TimerSink) then

                        -- Sink object for 5 seconds
                        local ease = self.SpeedFall
                        if self.TimerList[i] > (self.TimerSink - 5) then
                            ease = (self.SpeedFall / (self.TimerList[i] - 5))
                        end

                        local move = dt * self.SpeedFall * ease
                        obj:move({0, -move, 0})

                        -- Clear Ground
                        if not self.ClearedList[i] then
                            self:getLevel():createObject("PREFAB_FIRE_CLEARING", obj:getGlobalPosition())
                            self.ClearedList[i] = true
                        end


                    -- Object has spread, has sunk, time has elapsed to allow emitters to finish, and all are ready to destroy
                    elseif self.StateList[i] == 2 and self.TimerList[i] < (0 - self.TimerSink) then

                        -- Don't destroy a fire starter as this causes a crash and will stop the fire from working
                        if obj.Name == "FireStarter.Part" then
                            self.StateList[i] = 3 -- sets this outside the active list of objects but does not destroy it

                            self.ActiveCount = self.ActiveCount - 1

                            if self.EmitterList[i] then
                                local objEmitter = self:getLevel():find(self.EmitterList[i])
                                if objEmitter then
                                    objEmitter:destroy()
                                end
                            end


                        else

                            -- If this is the only building part, delete the entire building (otherwise you get a crash);
                            -- if a part within a larger monument, just delete this part
                            local compBuilding = obj:findFirstObjectWithComponentUp("COMP_BUILDING")
                            if compBuilding then

                                -- Get count of building parts
                                local listCompPart = compBuilding:getBuildingPartList()
                                if listCompPart then

                                    if #listCompPart == 1 then

                                        -- Delete the entire building
                                        local objBuilding = compBuilding:getOwner()
                                        if objBuilding then
                                            -- mod:log("COMP_ORIGIN is deleting the entire building: " .. objBuilding.Name)
                                            if compBuilding.EditMode then
                                                compBuilding.EditMode = false
                                            end
                                            objBuilding:destroy()
                                        end

                                    else

                                        -- Delete only this part
                                        -- mod:log("COMP_ORIGIN is deleting just one part: " .. objLinked.Name)
                                        local compPart = obj:getComponent("COMP_BUILDING_PART")
                                        if compPart then
                                            if compPart:getIsBuilt() then -- do not delete parts under construction, causes a crash
                                                obj:destroy()
                                            end
                                        end

                                    end
                                end

                            else

                                -- Not a building, just delete the object directly
                                obj:destroy()

                            end -- check if building


                            -- Promote to final state
                            self.StateList[i] = 3
                            self.ActiveCount = self.ActiveCount - 1


                            -- Remove any spawned emitters
                            if self.EmitterList[i] then
                                local objEmitter = self:getLevel():find(self.EmitterList[i])
                                if objEmitter then
                                    objEmitter:destroy()
                                end
                            end


                        end -- not a fire starter

                    end -- state and timer checks


                -- Object no longer exists, change status
                else
                    self.StateList[i] = 3
                    self.ActiveCount = self.ActiveCount - 1


                end -- object exists

            end -- state >

        end -- loop through all objects


        -- Finished fighting this fire
        -- mod:log("ActiveCount " .. self.ActiveCount .. ", WaterCount " .. self.WaterCount)

        if self.ActiveCount == 0 then

            -- Unregister the fire with the controller
            local compControl = self:getLevel():getComponentManager("COMP_FIRE_CONTROLLER"):getFirstEnabled()
            if compControl then
                compControl:unregisterFire(self:getOwner())
            end


            -- Remove all spawned emitters that might remain
            for j = 1, #self.EmitterList do
                if self.EmitterList[j] then
                    local objEmitter = self:getLevel():find(self.EmitterList[j])
                    if objEmitter then
                        objEmitter:destroy()
                    end
                end
            end


            -- Delete this fire
            self:getOwner():destroy()


        end -- there are no actively burning objects left

    end -- fps limiter

end -- COMP_FIRE update()



-- Set object on fire -------------------------------------------------------------------------------------
function COMP_FIRE:setObjectOnFire(obj, i)

    local posObj = obj:getGlobalPosition()

    -- Objects with extreme scales will create problems with the particle effect, skip them
    if obj.Scale[2] > 0.2 and obj.Scale[2] < 5.0 then

        -- Set state
        self.StateList[i] = 1
        self.ActiveCount = self.ActiveCount + 1

        -- Put emitter on the object
        local objEmitter = self:getLevel():createObject("PREFAB_FIRE_EMITTER", posObj)
        -- mod:log("Emitter created at " .. tostring(posObj))
        self.EmitterList[i] = objEmitter:getId()

        -- Add emitters to child objects
        local childLimit = self.ChildEmitterLimit
        obj:forEachChild(function(child)

            if childLimit > 0 then

                if child.Scale[2] > 0.2 and child.Scale[2] < 5 and childLimit > 0 then -- objects with extreme scales will create problems with the particle effect

                    local posChild = child:getGlobalPosition()
                    local dist = math.sqrt( (posObj[1] - posChild[1])^2 + (posObj[3] - posChild[3])^2 )

                    if (dist > 1) and (dist < self.EachSpreadDistance) then -- make sure the child is far enough away but not too far

                        -- Put emitter on some children (not all due to performance)
                        local objChildEmitter = self:getLevel():createObject("PREFAB_FIRE_EMITTER", posChild)
                        -- mod:log("Child emitter at " .. tostring(posChild))
                        objChildEmitter:setParent(objEmitter, true)
                        childLimit = childLimit - 1

                    end

                end

            end

        end)

    end

end



-- Register
mod:registerClass(COMP_FIRE)

-- Attach component
mod:registerPrefabComponent("PREFAB_FIRE_ORIGIN", { DataType = "COMP_FIRE" } ) -- used by other mods like the dragon, or to create fire spontaneously
mod:registerPrefabComponent("PREFAB_FIRE_START", { DataType = "COMP_FIRE" } ) -- used by the player placing the fire manually


